from abc import ABC, abstractmethod
from typing import Callable, Any

from cloudevents.http import CloudEvent


class Processor(ABC):

    @abstractmethod
    def mapper(self) -> Callable[[CloudEvent], Any]:
        pass

    @abstractmethod
    def close(self) -> None:
        pass



import logging
from logging import Logger
from typing import Callable, Any, cast

from cloudevents.http import CloudEvent

from .processor import Processor


class LoggingProcessor(Processor):

    def __init__(self, logger: Logger, level: int):
        self.__logger: Logger = logger
        self.__level: int = level

    def __str__(self) -> str:
        return "logging"

    def mapper(self) -> Callable[[CloudEvent], Any]:
        return lambda event: self.__logger.log(level=self.__level, msg=event)

    def close(self) -> None:
        self.__logger.log(level=logging.INFO, msg='closing resources...')


def logging_processor_builder(**kwargs: dict) -> LoggingProcessor:
    if kwargs.get('logger') is None or not hasattr(kwargs.get('logger'), 'log') \
            or kwargs.get('level') is None or not isinstance(kwargs.get('level'), int):
        raise ValueError
    # noinspection PyTypeChecker,PyDeepBugsSwappedArgs
    return LoggingProcessor(cast(Logger, kwargs.get('logger')), cast(int, kwargs.get('level')))



from typing import Callable, Any, cast

from cloudevents.http import CloudEvent

from .history import History
from .processor import Processor


class HistoryProcessor(Processor):

    def __init__(self, history: History):
        self.__history = history

    def __str__(self) -> str:
        return "history"

    def mapper(self) -> Callable[[CloudEvent], Any]:
        return lambda x: self.__history.persist(str(x))

    def close(self) -> None:
        self.__history.close()


def history_processor_builder(**kwargs: dict) -> HistoryProcessor:
    if kwargs.get('history') is None or not hasattr(kwargs.get('history'), 'persist'):
        raise ValueError
    # noinspection PyTypeChecker
    return HistoryProcessor(cast(History, kwargs.get('history')))



from collections import Callable
from typing import List, Tuple, cast

from .fluentd_processor import fluentd_processor_builder
from .history_processor import history_processor_builder
from .logging_processor import logging_processor_builder
from .processor import Processor


def get_builder_map(**kwargs: dict) -> dict[str, dict]:
    return {
        'history': {
            'builder': history_processor_builder,
            'kwargs': {
                'history': kwargs.get('history')
            }
        },
        'logging': {
            'builder': logging_processor_builder,
            'kwargs': {
                'logger': kwargs.get('logger'),
                'level': kwargs.get('logger.level')
            }
        },
        'fluentd': {
            'builder': fluentd_processor_builder,
            'kwargs': {
                'fluentd': kwargs.get('fluentd')
            }
        }
    }


def processors_builder(builder_map: Tuple[dict]) -> List[Processor]:
    result: List[Processor] = []
    for item in builder_map:
        proc_i: Callable = cast(Callable, item.get('builder'))
        # noinspection PyDeepBugsSwappedArgs
        kwargs_i: dict = cast(dict, item.get('kwargs'))
        result.append(proc_i(**kwargs_i))
    return result



from typing import Dict, Any, Tuple, Text, List

from cloudevents.http import CloudEvent

from .processor import Processor
from ..common.config import Config


def config2attributes(config: Config) -> Any:
    return config.get_section('CloudEvents')


def get_configuration_list(config: Config) -> Tuple[Text, ...]:
    return config.get_tuple(section='Reports', key='processors')


class Report(object):
    def __init__(self, attributes: dict, processors: List[Processor]):
        self.__attributes: Dict = attributes
        self.__processors: List[Processor] = processors

    def add_event(self, record: dict) -> None:
        event: CloudEvent = CloudEvent(attributes=self.__attributes, data=record)
        for x in self.__processors:
            x.mapper()(event)

    def close(self) -> None:
        for proc in self.__processors:
            proc.close()



from typing import Callable, Any, cast

from cloudevents.http import CloudEvent

from .fluentd_logger import FluentdLogger
from .processor import Processor


class FluentdProcessor(Processor):

    def __init__(self, fluentd_logger: FluentdLogger):
        self.__fluentd_logger = fluentd_logger

    def mapper(self) -> Callable[[CloudEvent], Any]:
        return lambda x: self.__fluentd_logger.forward(x)

    def close(self) -> None:
        self.__fluentd_logger.close()


def fluentd_processor_builder(**kwargs: dict) -> FluentdProcessor:
    if kwargs.get('fluentd') is None or not hasattr(kwargs.get('fluentd'), 'forward'):
        raise ValueError
    # noinspection PyTypeChecker
    return FluentdProcessor(cast(FluentdLogger, kwargs.get('fluentd')))



import logging.config
from logging import Logger


# noinspection StrFormat
def logger_setup(log_cfg: dict) -> None:
    logging.config.dictConfig(log_cfg)


def get_logger(name: str) -> Logger:
    return logging.getLogger(name=name)



import json
import logging

from cloudevents.http import CloudEvent, to_json
from fluent import sender


class FluentdLogger(object):

    def __init__(self, tag: str, label: str) -> None:
        self.__sender = sender.FluentSender(tag=tag, host='127.0.0.1', port=24224, nanosecond_precision=True)
        self.__label = label

    def forward(self, event: CloudEvent) -> None:
        data: dict = json.loads(to_json(event))
        if not self.__sender.emit(label=self.__label, data=data):
            logging.error(self.__sender.last_error)
            self.__sender.clear_last_error()

    def close(self) -> None:
        self.__sender.close()
